# -*- coding: utf-8 -*-
"""
Performs dimensionality reduction on a Space object.

The Space object created by reading in the sparse co-occurrence matrix has a very high dimensionality. Because the true
number of latent dimensions is thought to be much lower, we need to reduce the number of dimensions. This function
uses two different methods and writes the results to disk. Because the ideal number of dimensions may not be known
ahead of time, the method expects a range and a step.
@author: Max
"""

import os
from composes.transformation.dim_reduction.svd import Svd
from composes.transformation.dim_reduction.nmf import Nmf
from composes.semantic_space.space import Space
from composes.utils import io_utils
import pickle
import argparse
import sys


def main(argv):
    parser = argparse.ArgumentParser(description='Performs dimensionality reduction on a Space object.')
    parser.add_argument('-m', action='store', dest='spacefile',
                        help='Input the name of the model file', required=True)
    parser.add_argument('-n', action='store', dest='dmin',
                        help='The min. # of dimensions to reduce the model to', default=50, type=int)
    parser.add_argument('-x', action='store', dest='dmax',
                        help='The max. # of dimensions to reduce the model to', default=150, type=int)
    parser.add_argument('-s', action='store', dest='step',
                        help='The step-size to increment from min. to max.', default=15, type=int)
    parser.add_argument('-d', action='store', dest='dimRedMethod', help='Dimensionality Reduction method. 0=SVD, 1=NMF',
                        default=0, type=int)

    args = vars(parser.parse_args())
    spacefile = args['spacefile']
    dmax = args['dmax']
    dmin = args['dmin']
    step = args['step']
    dimRedMethod = args['dimRedMethod']

    # load the input file
    f = open(spacefile, 'rb')
    my_space = pickle.load(f)

    # perform truncated SVD on the input file for dimensions ranging from dmin to dmax
    curNumDims = dmin
    while curNumDims <= dmax:
        if dimRedMethod == 0:
            svd_reduced_space = my_space.apply(Svd(curNumDims))
            io_utils.save(svd_reduced_space, './' + str(spacefile) + '_SVD' + str(curNumDims) + '.pkl')
        elif dimRedMethod == 1:
            nmf_reduced_space = my_space.apply(Nmf(curNumDims))
            io_utils.save(nmf_reduced_space, './' + str(spacefile) + '_NMF' + str(curNumDims) + '.pkl')
        curNumDims += step


if __name__ == "__main__":
    main(sys.argv[1:])
